Imports System.IO
Imports System.Text
Imports System.Xml
Imports iTextSharp
Imports System.ComponentModel
Imports System.Reflection

Namespace WESNet.PDF

    'The PDFGrid class defines a section that is rendered much like a DataGrid or DataView control on
    'a web page. In fact, its XML definition is similar to that of the DataGrid.

    Public Class PDFGrid
        Inherits BasePDFList

#Region "Private Members"
        Private _AutoGenerateColumns As Boolean = True
        Private _Columns As PDFGridColumnCollection
        Private _Items As PDFItemCollection
        Private _Widths As Single()             'Array of column widths

        Private Const DefaultGridWidth As Single = 0.5!

#End Region

#Region "Public Properties"

        'If true (default if true), columns will be auto-generated from the datasource schema
        Public Property AutoGenerateColumns() As Boolean
            Get
                Return _AutoGenerateColumns
            End Get
            Set(ByVal Value As Boolean)
                _AutoGenerateColumns = Value
            End Set
        End Property

        'A collection of PDFGridColumn objects representing the columns of the PDFGrid
        Public ReadOnly Property Columns() As WESNet.PDF.PDFGridColumnCollection
            Get
                If _Columns Is Nothing Then
                    _Columns = New PDFGridColumnCollection(Me)
                End If
                Return _Columns
            End Get
        End Property

        'Array containing the widths of each column
        Public Property Widths() As Single()
            Get
                Return _Widths
            End Get
            Set(ByVal Value As Single())
                _Widths = Value
            End Set
        End Property

        'A collection of PDFItems representing the rows of the PDFGrid
        Public ReadOnly Property Items() As WESNet.PDF.PDFItemCollection
            Get
                If _Items Is Nothing Then
                    _Items = New PDFItemCollection(Me)
                End If
                Return _Items
            End Get
        End Property

#End Region

#Region "Private/Protected/Friend Methods"
        Protected Overrides Sub OnDataBinding(ByVal e As EventArgs)
            Dim enumerator As IEnumerator
            Dim HasDTForDVF, HasDTFS As Boolean
            Dim DKF As String = Me.DataKeyField
            Dim HasDKF As Boolean = (Not DKF Is Nothing) AndAlso (DKF.Length > 0)
            Dim DKFDescriptor As PropertyDescriptor
            Dim DataIdx As Integer = -1
            Dim ItemIdx As Integer = 0
            Dim ItemType As System.Web.UI.WebControls.ListItemType
            Dim DataRowObj As Object
            Dim Finished As Boolean

            MyBase.OnDataBinding(e)
            With PagedDataSource
                .PageSize = Me.RecordsPerPage
                .AllowPaging = Me.EnablePaging
                .AllowCustomPaging = False
                .VirtualCount = 0
                .DataSource = GetResolvedDataSource()
            End With

            If Not PagedDataSource.DataSource Is Nothing Then
                If Not _Items Is Nothing Then Items.Clear()
                DataKeysArray.Clear()

                Dim GridBorder As New Border

                Select Case GridLines
                    Case Web.UI.WebControls.GridLines.None
                        GridBorder = Border.Empty
                    Case Web.UI.WebControls.GridLines.Horizontal
                        GridBorder.Bottom = New Edge(Unit.Point(DefaultGridWidth), Web.UI.WebControls.BorderStyle.Solid, System.Drawing.Color.Black)
                        GridBorder.Top = New Edge(GridBorder.Bottom)
                    Case Web.UI.WebControls.GridLines.Vertical
                        GridBorder.Right = New Edge(Unit.Point(DefaultGridWidth), Web.UI.WebControls.BorderStyle.Solid, System.Drawing.Color.Black)
                        GridBorder.Left = New Edge(GridBorder.Right)
                    Case Web.UI.WebControls.GridLines.Both
                        GridBorder = New Border(Unit.Point(DefaultGridWidth), Web.UI.WebControls.BorderStyle.Solid, System.Drawing.Color.Black)
                End Select
                HeaderStyle.Border.MergeWith(GridBorder)
                ItemStyle.Border.MergeWith(GridBorder)
                AlternatingItemStyle.MergeWith(ItemStyle)
                FooterStyle.MergeWith(HeaderStyle)
                PagerStyle.MergeWith(FooterStyle)

                Do
                    PagedDataSource.CurrentPageIndex = Me.CurrentPageIndex
                    'Get the first row of data (necessary to auto-generate columns with Field information
                    enumerator = PagedDataSource.GetEnumerator

                    Finished = Not enumerator.MoveNext()
                    If Not Finished Then
                        DataRowObj = enumerator.Current
                        If PagedDataSource.IsFirstPage Then
                            CreateColumnSet(DataRowObj)
                            If HasDKF Then
                                DKFDescriptor = TypeDescriptor.GetProperties(DataRowObj).Find(DKF, True)
                                If DKFDescriptor Is Nothing Then
                                    Throw New DataException("Data Key Field " & DKF & " not found in data source")
                                End If
                            End If
                        Else
                            'Force a new page

                        End If
                        If ShowHeader Then
                            CreateItem(ItemIdx, -1, Web.UI.WebControls.ListItemType.Header, False, Nothing)
                        End If
                    End If
                    Do Until Finished
                        DataRowObj = enumerator.Current
                        DataIdx += 1
                        ItemIdx += 1
                        If ItemIdx Mod 2 = 1 Then
                            ItemType = Web.UI.WebControls.ListItemType.Item
                        Else
                            ItemType = Web.UI.WebControls.ListItemType.AlternatingItem
                        End If
                        CreateItem(ItemIdx, DataIdx, ItemType, True, DataRowObj)
                        If Not DKFDescriptor Is Nothing Then
                            DataKeysArray.Add(DKFDescriptor.GetValue(DataRowObj))
                        End If
                        Finished = Not enumerator.MoveNext
                    Loop
                    If TypeOf DataSource Is IDataReader Then
                        Dim dr As IDataReader = CType(DataSource, IDataReader)
                        If Not dr Is Nothing AndAlso Not dr.IsClosed Then
                            dr.Close()
                        End If
                    End If
                    If ShowFooter Then
                        CreateItem(ItemIdx, -1, Web.UI.WebControls.ListItemType.Footer, False, Nothing)
                    End If

                    If EnablePaging Then
                        CreateItem(ItemIdx, -1, Web.UI.WebControls.ListItemType.Pager, False, Nothing)
                    End If

                    CurrentPageIndex += 1
                Loop Until PagedDataSource.IsLastPage
            End If
        End Sub

        Protected Sub ParseColumns(ByVal Node As XmlNode)
            Dim ColumnsNode As XmlNode = Node.Item("Columns")
            If Not ColumnsNode Is Nothing Then
                AutoGenerateColumns = False
                Dim ChildNodes As XmlNodeList = ColumnsNode.ChildNodes
                For Each ChildNode As XmlNode In ChildNodes
                    If ChildNode.Name <> "Column" Then
                        Throw New Exception("Columns node must include one or more Column nodes")
                    End If
                    Dim Col As New WESNet.PDF.PDFGridColumn

                    With Col
                        .HeaderText = CType(Common.GetNamedAttribute(ChildNode, "HeaderText", String.Empty), String)
                        .FooterText = CType(Common.GetNamedAttribute(ChildNode, "FooterText", String.Empty), String)
                        .DataField = CType(Common.GetNamedAttribute(ChildNode, "DataField", String.Empty), String)
                        .DataFormatString = CType(Common.GetNamedAttribute(ChildNode, "DataFormatString", String.Empty), String)
                        Dim PaddingString As String = CType(Common.GetNamedAttribute(ChildNode, "Padding", String.Empty), String)
                        Col.Padding = QuadUnit.Parse(PaddingString)
                    End With
                    Columns.Add(Col)

                    If ChildNode.HasChildNodes Then
                        Dim ColumnStyles As XmlNodeList = ChildNode.ChildNodes
                        For Each ColumnStyle As XmlNode In ColumnStyles
                            Select Case ColumnStyle.Name
                                Case "HeaderStyle", "ItemStyle", "AlternatingItemStyle", "FooterStyle"
                                    Common.SetSubProperties(Col, ColumnStyle.Name, ColumnStyle)
                                Case Else
                                    Throw New XmlException("Unrecognized child node (" & ColumnStyle.Name & ") in column definition (" & Col.DataField & ")")
                            End Select
                        Next
                    End If
                Next
            End If
        End Sub

        Private Sub CreateColumnSet(ByVal DataRowObj As Object)
            If Not DataRowObj Is Nothing Then
                If Me.AutoGenerateColumns Then
                    CreateAutoGeneratedColumns(DataRowObj)
                Else
                    For Each Col As PDFGridColumn In Columns
                        Dim FieldDescriptor As PropertyDescriptor
                        FieldDescriptor = TypeDescriptor.GetProperties(DataRowObj).Find(Col.DataField, True)
                        With Col
                            .FieldDescriptor = FieldDescriptor
                            .HeaderStyle.MergeWith(HeaderStyle)
                            .FooterStyle.MergeWith(FooterStyle)
                            .ItemStyle.MergeWith(ItemStyle)
                            .AlternatingItemStyle.MergeWith(AlternatingItemStyle)
                            .AlternatingItemStyle.MergeWith(Col.ItemStyle)
                        End With
                    Next
                End If
            End If
        End Sub

        Private Sub CreateAutoGeneratedColumns(ByVal DataRowObj As Object)
            Dim PropertyDescriptors As PropertyDescriptorCollection
            Dim FieldDescriptor As PropertyDescriptor
            Dim Col As WESNet.PDF.PDFGridColumn
            If Not DataRowObj Is Nothing Then
                PropertyDescriptors = TypeDescriptor.GetProperties(DataRowObj)
                If PropertyDescriptors Is Nothing Then
                    Columns.Add("", "Item")
                Else
                    For Each FieldDescriptor In PropertyDescriptors
                        Col = New WESNet.PDF.PDFGridColumn(FieldDescriptor.Name)
                        Col.FieldDescriptor = FieldDescriptor
                        Col.HeaderStyle.CopyFrom(HeaderStyle)
                        Col.FooterStyle.CopyFrom(FooterStyle)
                        Col.ItemStyle.CopyFrom(ItemStyle)
                        Col.AlternatingItemStyle.CopyFrom(AlternatingItemStyle)
                        Columns.Add(Col)
                    Next
                End If
            End If
        End Sub

        Private Function CreateItem(ByVal ItemIdx As Integer, ByVal DataIdx As Integer, ByVal ItemType As System.Web.UI.WebControls.ListItemType, ByVal DataBinding As Boolean, _
                                    ByVal DataRowObj As Object) As PDFItem

            Dim Cell As WESNet.PDF.TableCell
            Dim i As Integer
            Dim Item As PDFItem = New PDFItem(ItemIdx, DataIdx, ItemType)
            Dim Args As New PDFItemEventArgs(Item)

            If (ItemType <> System.Web.UI.WebControls.ListItemType.Pager) Then
                For i = 0 To Columns.Count - 1
                    Cell = New WESNet.PDF.TableCell(Item)
                    Select Case ItemType
                        Case Web.UI.WebControls.ListItemType.Header
                            Cell.Text = Columns.Item(i).HeaderText
                            Cell.Style.CopyFrom(Columns.Item(i).HeaderStyle)
                        Case Web.UI.WebControls.ListItemType.Footer
                            Cell.Text = Columns.Item(i).FooterText
                            Cell.Style.CopyFrom(Columns.Item(i).FooterStyle)
                        Case Web.UI.WebControls.ListItemType.AlternatingItem
                            Cell.Style.CopyFrom(Columns.Item(i).AlternatingItemStyle)
                        Case Web.UI.WebControls.ListItemType.Item
                            Cell.Style.CopyFrom(Columns.Item(i).ItemStyle)
                        Case Else
                            Throw New Exception("Invalid ListItemType = " & ItemType.ToString())
                    End Select
                    Item.Cells.Add(Cell)
                Next
                Items.Add(Item)
                OnItemCreated(Args)
                If DataBinding Then
                    Item.DataItem = DataRowObj
                    For i = 0 To Columns.Count - 1
                        Item.Cells.Item(i).Text = Columns.Item(i).GetFormattedValue(DataRowObj)
                    Next
                    OnItemDataBound(Args)
                    Item.DataItem = Nothing
                End If
            Else 'Item is Pager
                Cell = New WESNet.PDF.TableCell(Item)
                Cell.ColumnSpan = Columns.Count
                Cell.HorizontalAlign = Web.UI.WebControls.HorizontalAlign.Right
                Cell.Text = "Page " & (PagedDataSource.CurrentPageIndex + 1).ToString & " of " & _
                                       (PagedDataSource.PageCount).ToString
                Cell.Style.CopyFrom(PagerStyle)
                Me.OnItemCreated(Args)
                Item.Cells.Add(Cell)
                Items.Add(Item)
                OnItemCreated(Args)
            End If
            Return Item
        End Function

#End Region

#Region "Constructors"
        Public Sub New()
            MyBase.New(SectionTypes.PDFGrid)
        End Sub

        Public Sub New(ByVal Node As XmlNode)
            Me.New()
            ParseSectionAttributes(Node)
            ParseSectionStyle(Node)
            ParseDataSource(Node)
            ParsePDFListStyles(Node)
            ParseColumns(Node)
        End Sub

#End Region




        Private Function CreateTable() As iTextSharp.text.Table
            Dim Table As New iTextSharp.text.Table(Columns.Count)
            With Table
                If Not Width.IsEmpty Then
                    If Width.Type = Web.UI.WebControls.UnitType.Percentage Then
                        .WidthPercentage = Width.Value
                    Else
                        'This does not work - cannot set width other than by percentage??
                        .AbsWidth = Width.ToPoints.ToString
                    End If
                End If
                .Alignment = Common.TranslateHorizontalAlignment(HorizontalAlign)
                Common.SetColor(.BackgroundColor, BackColor)
                If BorderStyle <> Web.UI.WebControls.BorderStyle.None Then
                    .Border = iTextSharp.text.Table.BOX
                    .BorderWidth = BorderWidth.ToPoints()
                    Common.SetColor(.BorderColor, BorderColor)
                End If
                If Not CellPadding.IsEmpty Then .Cellpadding = CellPadding.ToPoints()
                If CellSpacing.IsEmpty OrElse CellSpacing.ToPoints = 0.0! Then
                    .Cellspacing = -DefaultGridWidth
                Else
                    .Cellspacing = CellSpacing.ToPoints() / 2.0!
                End If
            End With
            Return Table
        End Function


        Public Overrides Function RenderedContent(ByVal writer As PDFWriter) As iTextSharp.text.IElement
            Dim i As Integer
            Dim Item As PDFItem
            Dim Paragraph As New iTextSharp.text.Paragraph
NextTable:
            Dim Table As iTextSharp.text.Table = CreateTable()

            Do
                Item = Items.Item(i)
                i += 1
                If Item.ItemType = Web.UI.WebControls.ListItemType.Pager Then
                    Exit Do
                Else
                    Item.AddToTable(Table)
                End If
            Loop While i < Items.Count
            'Readjust column widths
            Widths = GetMaxColumnWidths(Table)
            Table.Widths = Widths
            If Item.ItemType <> Web.UI.WebControls.ListItemType.Pager Then
                Paragraph.Add(Table)
            Else
                If ShowPager Then Item.AddToTable(Table)
                Paragraph.Add(Table)
                If i < Items.Count Then
                    Paragraph.Add(iTextSharp.text.Chunk.NEXTPAGE)
                    GoTo NextTable
                End If
            End If
            Return Paragraph
        End Function

        Public Function GetMaxColumnWidths(ByVal Table As iTextSharp.text.Table) As Single()
            Dim Row, Col, RowIdx, ColIdx As Integer
            Dim Rows As Integer = Table.Size
            Dim Cols As Integer = Table.Columns
            Dim TmpWidths(Cols - 1) As Single
            Dim TableElement As Object
            Dim Cell As iTextSharp.text.Cell
            Dim Chunk As iTextSharp.text.Chunk
            Dim Width As Single
            Dim TotalWidth As Single
            Dim TotalPercentage As Single

            For RowIdx = 0 To Rows - 1
                For ColIdx = 0 To Cols - 1
                    TableElement = Table.GetElement(RowIdx, ColIdx)
                    If TypeOf (TableElement) Is iTextSharp.text.Cell Then
                        Cell = CType(TableElement, iTextSharp.text.Cell)
                        If Cell.Chunks.Count > 0 Then
                            Chunk = CType(Cell.Chunks(0), iTextSharp.text.Chunk)
                            Width = Chunk.GetWidthPoint()
                            Width = Width + (2.0! * Table.Cellpadding) + Table.Cellspacing
                            If Width > TmpWidths(ColIdx) Then TmpWidths(ColIdx) = Width
                        End If
                    ElseIf TypeOf (TableElement) Is iTextSharp.text.Table Then
                        Throw New ApplicationException("Cannot handle table cells that are tables at this time")
                    End If
                Next
            Next
            For ColIdx = 0 To Cols - 1
                TotalWidth = TotalWidth + TmpWidths(ColIdx)
            Next
            TotalWidth = TotalWidth + 2.0! * Table.BorderWidth
            For ColIdx = 0 To Cols - 1
                TmpWidths(ColIdx) = (TmpWidths(ColIdx) / TotalWidth) * 100.0!
                TotalPercentage = TotalPercentage + TmpWidths(ColIdx)
            Next
            TmpWidths(Cols - 1) = TmpWidths(Cols - 1) + (100.0! - TotalPercentage)
            Return TmpWidths
        End Function

    End Class


End Namespace
