Imports iTextSharp
Imports iTextSharp.text
Imports System.IO

Namespace WESNet.PDF

    'Extends the iTextSharp PDFPageEventHelper class to handle the additional properties and
    'capabilities encapsulated in the PDFGenerator object and is used to render the PDF report

    Public Class PDFWriter
        Inherits iTextSharp.text.pdf.PdfPageEventHelper

#Region "Private Members"
        Private _PDFDoc As Document
        Private _PDFStream As Stream
        Private _PDFWriter As iTextSharp.text.pdf.PdfWriter
        Private _LastElementAdded As IElement
        Private _Title As String = String.Empty
        Private _Author As String = String.Empty
        Private _Creator As String = String.Empty
        Private _DateCreated As DateTime = Now()
        Private _Subject As String = String.Empty
        Private _PageSize As iTextSharp.text.Rectangle = iTextSharp.text.PageSize.LETTER
        Private _PageColor As System.Drawing.Color = System.Drawing.Color.Empty
        Private _Orientation As Orientation = PDF.Orientation.Portrait
        Private _Margins As QuadUnit = New QuadUnit(Unit.Inch(0.5))
        Private _TotalPages As Integer
        Private _PageHeader As HeaderFooter
        Private _PageFooter As HeaderFooter
        Private _Tokens As Hashtable
#End Region

#Region "Private Constants"
        Private InitialBufferSize As Integer = 2096
#End Region


#Region "Public Properties"

        'Provides access to the underlying iTextSharp.text.Document class
        Public ReadOnly Property PDFDoc() As Document
            Get
                Return _PDFDoc
            End Get
        End Property

        'True if the PDFDoc has been opened
        Public ReadOnly Property DocumentIsOpen() As Boolean
            Get
                If _PDFDoc Is Nothing Then
                    Return False
                Else
                    Return _PDFDoc.IsOpen
                End If
            End Get
        End Property

        'Returns the I/O Stream of the PDF report being written
        Public ReadOnly Property PDFStream() As Stream
            Get
                Return _PDFStream
            End Get
        End Property

        'Provides access to the underlying iTextSharp.text.pdf.PdfWriter class
        Public ReadOnly Property PDFWriter() As iTextSharp.text.pdf.PdfWriter
            Get
                Return _PDFWriter
            End Get
        End Property

        'The report Title - cannot be set/changed once the PDF document is open
        Public Property Title() As String
            Get
                Return _Title
            End Get
            Set(ByVal Value As String)
                _Title = Value
                If DocumentIsOpen Then Throw New NotSupportedException("Title cannot be set after opening PDF document")
                _PDFDoc.AddTitle(_Title)
            End Set
        End Property

        'The report Author - cannot be set/changed once the PDF document is open
        Public Property Author() As String
            Get
                Return _Author
            End Get
            Set(ByVal Value As String)
                _Author = Value
                If DocumentIsOpen Then Throw New NotSupportedException("Author cannot be set after opening PDF document")
                _PDFDoc.AddAuthor(_Title)
            End Set
        End Property

        'The report creator - cannot be set/changed once the PDF document is open
        Public Property Creator() As String
            Get
                Return _Creator
            End Get
            Set(ByVal Value As String)
                _Creator = Value
                If DocumentIsOpen Then Throw New NotSupportedException("Creator cannot be set after opening PDF document")
                _PDFDoc.AddCreator(_Creator)
            End Set
        End Property

        'The report DateCreated - cannot be set/changed once the PDF document is open
        Public Property DateCreated() As DateTime
            Get
                Return _DateCreated
            End Get
            Set(ByVal Value As DateTime)
                _DateCreated = Value
                If DocumentIsOpen Then Throw New NotSupportedException("Date created cannot be set after opening PDF document")
                _PDFDoc.AddCreationDate()
            End Set
        End Property

        'The report Subject - cannot be set/changed once the PDF document is open
        Public Property Subject() As String
            Get
                Return _Subject
            End Get
            Set(ByVal Value As String)
                _Subject = Value
                If DocumentIsOpen Then Throw New NotSupportedException("Subject cannot be set after opening PDF document")
                _PDFDoc.AddSubject(Value)
            End Set
        End Property

        'The PageSize as an iTextSharp.text.Rectangle object
        Public Property PageSize() As iTextSharp.text.Rectangle
            Get
                Return _PageSize
            End Get
            Set(ByVal Value As iTextSharp.text.Rectangle)
                Dim OldValue As iTextSharp.text.Rectangle = _PageSize
                _PageSize = Value
                If Not Value Is OldValue Then OnPageSizeChanged()
            End Set
        End Property

        'The report Orientation - Landscape or Portrait
        Public Property Orientation() As PDF.Orientation
            Get
                Return _Orientation
            End Get
            Set(ByVal Value As PDF.Orientation)
                Dim OldValue As PDF.Orientation = _Orientation
                _Orientation = Value
                If Value <> OldValue Then OnOrientationChanged()
            End Set
        End Property

        'The report (background) PageColor
        Public Property PageColor() As System.Drawing.Color
            Get
                Return _PageColor
            End Get
            Set(ByVal Value As System.Drawing.Color)
                Dim OldValue As System.Drawing.Color = _PageColor
                _PageColor = Value
                If Not Value.Equals(OldValue) Then OnPageColorChanged()
            End Set
        End Property

        'The report margins as a QuadUnit - Top, Right, Bottom, Left
        Public Property Margins() As QuadUnit
            Get
                Return _Margins
            End Get
            Set(ByVal Value As QuadUnit)
                Dim OldValue As QuadUnit = _Margins
                _Margins = Value
                If Not Value.Equals(OldValue) Then OnMarginsChanged()
            End Set
        End Property

        'Provides access to the last IElement added to the PDFWriter during rendering
        Public ReadOnly Property LastElementAdded() As IElement
            Get
                Return _LastElementAdded
            End Get
        End Property

        'Returns the CurrentPage number during rendering
        Public ReadOnly Property CurrentPage() As Integer
            Get
                Return PDFWriter.PageNumber
            End Get
        End Property

        'Returns the total pages of the report at the end of rendering
        Public ReadOnly Property TotalPages() As Integer
            Get
                Return _TotalPages
            End Get
        End Property

        'Provides access to the PageHeader object
        Public ReadOnly Property PageHeader() As HeaderFooter
            Get
                Return _PageHeader
            End Get
        End Property

        'Provides access to the PageFooter object
        Public ReadOnly Property PageFooter() As HeaderFooter
            Get
                Return _PageFooter
            End Get
        End Property

        'Provides access to a Hashtable of tokens in label sections that will be replaced
        'with their actual values during and/or after rendering
        Public ReadOnly Property Tokens() As Hashtable
            Get
                If _Tokens Is Nothing Then
                    _Tokens = New Hashtable
                End If
                Return _Tokens
            End Get
        End Property
#End Region

#Region "Public Methods"
        'Opens the current PDF document and initializes any PageFooter/PageHeader objects
        Public Overridable Sub OpenDocument()
            If Not PDFDoc Is Nothing Then
                Dim Footer As iTextSharp.text.HeaderFooter
                Dim Header As iTextSharp.text.HeaderFooter
                Dim Phrase As iTextSharp.text.Phrase
                If Not PageFooter Is Nothing Then
                    Phrase = CType(PageFooter.RenderedContent(Me), iTextSharp.text.Phrase)
                    If Not Phrase Is Nothing Then
                        Footer = New iTextSharp.text.HeaderFooter(Phrase, False)
                        Common.SetHorizontalAlignment(Footer.Alignment, PageFooter.Style.HorizontalAlign)
                        Common.SetBorders(Footer.Border, PageFooter.Style.Border)
                        PDFDoc.Footer = Footer
                    End If
                End If
                If Not PageHeader Is Nothing Then
                    Phrase = CType(PageHeader.RenderedContent(Me), iTextSharp.text.Phrase)
                    If Not Phrase Is Nothing Then
                        Header = New iTextSharp.text.HeaderFooter(Phrase, False)
                        Common.SetHorizontalAlignment(Header.Alignment, PageHeader.Style.HorizontalAlign)
                        Common.SetBorders(Header.Border, PageHeader.Style.Border)
                        PDFDoc.Header = Header
                    End If
                End If
                PDFDoc.Open()
            End If
        End Sub

        'Closes the PDF document - no further writing is possible
        Public Sub CloseDocument()
            If Not PDFDoc Is Nothing Then PDFDoc.Close()
        End Sub

        'Sends the PDF document to the client's browser via the HttpResponse stream.
        Public Sub StreamPDF(ByVal Response As System.Web.HttpResponse, ByVal OpenNewWindow As Boolean, ByVal OutputFilePath As String)
            If Not Response Is Nothing And Not PDFStream Is Nothing Then
                If DocumentIsOpen Then
                    PDFDoc.Close()
                End If
                PDFStream.Flush()
                PDFStream.Close()
                Dim Buffer As Byte() = CType(PDFStream, MemoryStream).ToArray()
                Dim BufLen As Integer = Buffer.Length
                Response.Clear()
                Try
                    Response.ClearHeaders()
                Catch ex As Exception
                    'Eat the exception which results from clearing the cookies collection but is
                    'OK in this case
                End Try

                'With Response
                '    .StatusCode = 200
                '    .Charset = Nothing
                '    .RedirectLocation = Nothing
                '    .SuppressContent = False
                '    .ExpiresAbsolute = DateTime.MinValue
                '    .CacheControl = Nothing
                'End With

                If OutputFilePath Is Nothing OrElse OutputFilePath = String.Empty Then
                    If Title Is Nothing OrElse Title = String.Empty Then
                        OutputFilePath = "Report.pdf"
                    Else
                        OutputFilePath = Title & ".pdf"
                    End If
                Else
                    If Path.GetExtension(OutputFilePath).ToLower <> ".pdf" Then
                        OutputFilePath = Path.GetFileNameWithoutExtension(OutputFilePath) & ".pdf"
                    End If
                End If
                Response.AppendHeader("Content-Length", BufLen.ToString)
                If OpenNewWindow Then
                    Response.AppendHeader("content-disposition", "attachment; filename=" & OutputFilePath)
                Else
                    Response.AppendHeader("content-disposition", "inline; filename=" & OutputFilePath)
                End If
                Response.AppendHeader("content-type", "application/pdf")
                Response.Flush() 'send the header information to the client
                Response.OutputStream.Write(Buffer, 0, BufLen)
                Response.Flush()
                Response.End()
            End If
        End Sub

        'Saves the PDF document to the specified filepath
        Public Sub SavePDF(ByVal Path As String)
            If Not PDFStream Is Nothing Then
                If DocumentIsOpen Then
                    PDFDoc.Close()
                End If
                PDFStream.Flush()
                PDFStream.Close()
                Dim FS As New FileStream(Path, FileMode.Create)
                Dim Buffer As Byte() = CType(PDFStream, MemoryStream).ToArray()
                Dim BufLen As Integer = Buffer.Length
                FS.Write(Buffer, 0, BufLen)
                FS.Flush()
                FS.Close()
            End If
        End Sub

        'Changes the PageSize - a new page is forced
        Public Sub ChangePageSize()
            ForceNewPage()
            If Orientation = Orientation.Landscape Then
                PDFDoc.SetPageSize(PageSize.Rotate())
            Else
                PDFDoc.SetPageSize(PageSize)
            End If
        End Sub

        'Forces a new page at this point in the PDF rendering
        Public Sub ForceNewPage()
            If DocumentIsOpen Then PDFDoc.NewPage()
        End Sub

        'Adds the given IElement to the PDF document
        Public Sub AddToDocument(ByVal Element As IElement)
            _LastElementAdded = Element
            PDFDoc.Add(Element)
        End Sub

        'Addes the given IElement to either the PDF document root or to the last element added
        Public Sub Add(ByVal Element As IElement)
            If LastElementAdded Is Nothing Then
                AddToDocument(Element)
            Else
                LastElementAdded.Chunks.Add(Element)
            End If
        End Sub

        Public Sub SetFooter(ByVal Footer As HeaderFooter)
            _PageFooter = Footer
        End Sub

        Public Sub SetHeader(ByVal Header As HeaderFooter)
            _PageHeader = Header
        End Sub
#End Region

#Region "Constructors"
        Public Sub New()
            _PDFDoc = New iTextSharp.text.Document
            _PDFStream = New System.IO.MemoryStream(InitialBufferSize)
            _PDFWriter = iTextSharp.text.pdf.PdfWriter.GetInstance(_PDFDoc, _PDFStream)
            _PDFWriter.PageEvent = Me
        End Sub

        Public Sub New(ByVal PDFDoc As iTextSharp.text.Document, ByVal PDFStream As System.IO.Stream)
            _PDFDoc = PDFDoc
            _PDFStream = PDFStream
            _PDFWriter = iTextSharp.text.pdf.PdfWriter.GetInstance(_PDFDoc, _PDFStream)
        End Sub

        Public Sub New(ByVal PageSize As iTextSharp.text.Rectangle)
            Me.New()
            _PageSize = PageSize
            PDFDoc.SetPageSize(PageSize)
        End Sub

        Public Sub New(ByVal PageSize As iTextSharp.text.Rectangle, ByVal Orientation As PDF.Orientation)
            Me.New()
            _PageSize = PageSize
            _Orientation = Orientation
            Dim Rect As iTextSharp.text.Rectangle = PageSize
            If Orientation = Orientation.Landscape Then
                Rect.Rotate()
            End If
            PDFDoc.SetPageSize(Rect)
        End Sub

        Public Sub New(ByVal PageSize As iTextSharp.text.Rectangle, ByVal Orientation As PDF.Orientation, ByVal PageColor As System.Drawing.Color)
            Me.New(PageSize, Orientation)
            _PageColor = PageColor
            PDFDoc.PageSize.BackgroundColor = Common.TranslateColor(PageColor)
        End Sub

        Public Sub New(ByVal PageSize As iTextSharp.text.Rectangle, ByVal Orientation As PDF.Orientation, _
                       ByVal PageColor As System.Drawing.Color, ByVal Margins As QuadUnit)
            Me.New(PageSize, Orientation, PageColor)
            _Margins = Margins
            PDFDoc.SetMargins(Margins.Left.ToPoints, Margins.Right.ToPoints, Margins.Top.ToPoints, Margins.Bottom.ToPoints)
        End Sub

#End Region

#Region "Events"
        Public Event PageSizeChanged As EventHandler
        Public Event OrientationChanged As EventHandler
        Public Event PageColorChanged As EventHandler
        Public Event MarginsChanged As EventHandler
#End Region

#Region "Event Handlers"

        'Is called when the PageSize is changed and raises the PageSizeChanged event
        Protected Overridable Sub OnPageSizeChanged()
            RaiseEvent PageSizeChanged(Me, New EventArgs)
            ChangePageSize()
        End Sub

        'Is called when the PageOrientation is changed and raises the OrientationChanged event
        Protected Overridable Sub OnOrientationChanged()
            RaiseEvent OrientationChanged(Me, New EventArgs)
            ChangePageSize()
        End Sub

        'Is called when the PageColor is changed and raises the PageColorChanged event
        Protected Overridable Sub OnPageColorChanged()
            RaiseEvent PageColorChanged(Me, New EventArgs)
            PDFDoc.PageSize.BackgroundColor = Common.TranslateColor(PageColor)
        End Sub

        'Is called when the Margins are changed and raises the MarginsChanged event
        Protected Overridable Sub OnMarginsChanged()
            RaiseEvent MarginsChanged(Me, New EventArgs)
            PDFDoc.SetMargins(Margins.Left.ToPoints, Margins.Right.ToPoints, Margins.Top.ToPoints, Margins.Bottom.ToPoints)
        End Sub


        'Is called when the PDF document is closed to update the TotalPages tokens with actual values
        Public Overrides Sub OnCloseDocument(ByVal writer As iTextSharp.text.pdf.PdfWriter, ByVal document As iTextSharp.text.Document)
            _TotalPages = writer.PageNumber - 1
            Dim Token As Token
            Dim TokenName As String
            For Each Entry As DictionaryEntry In Tokens
                Token = CType(Entry.Value, Token)
                TokenName = Token.Name
                Select Case TokenName
                    Case "TotalPages"
                        Token.SetValue(TotalPages)
                    Case "CurrentPage"
                        'Token.SetValue(CurrentPage)
                        'Do nothing as it has been handled in OnPageEnd
                    Case Else
                        Throw New iTextSharp.text.DocumentException("Unrecognized token " & TokenName)
                End Select
            Next
        End Sub

        'Is called at the end of each page's rendering to update CurrentPage tokens with actual values
        Public Overrides Sub OnEndPage(ByVal writer As iTextSharp.text.pdf.PdfWriter, ByVal document As iTextSharp.text.Document)
            Dim Token As Token
            Dim TokenName As String
            For Each Entry As DictionaryEntry In Tokens
                Token = CType(Entry.Value, Token)
                TokenName = Token.Name
                Select Case TokenName
                    Case "CurrentPage"
                        Token.SetValue(CurrentPage)
                    Case "TotalPages"
                        'Token.SetValue(TotalPages)
                        'Do nothing as it will be taken care of in OnCloseDocument
                    Case Else
                        Throw New iTextSharp.text.DocumentException("Unrecognized token " & TokenName)
                End Select
            Next

        End Sub

        'Is called during rendering of "OnGenericTags" to insert a TokenInstance Template into the PDF document
        Public Overrides Sub OnGenericTag(ByVal writer As iTextSharp.text.pdf.PdfWriter, ByVal document As iTextSharp.text.Document, ByVal rect As iTextSharp.text.Rectangle, ByVal text As String)
            Dim cb As iTextSharp.text.pdf.PdfContentByte = writer.DirectContent
            Dim TokenName As String() = text.Split(":"c)
            Dim Token As Token = CType(Tokens(TokenName(0)), Token)
            If Not Token Is Nothing Then
                Dim InstanceNumber As Integer = Integer.Parse(TokenName(1))
                Dim Instance As TokenInstance = CType(Token.Instances(InstanceNumber), TokenInstance)
                Instance.InsertTemplate(writer.DirectContent, rect, 0.0!)
            End If
        End Sub
#End Region

    End Class
End Namespace
