Imports iTextSharp
Imports System.Text
Imports System.Xml

Namespace WESNet.PDF

    'Represents a section of a PDF report and is the base class for all report elements
    Public MustInherit Class Section
        Inherits Object

#Region "Private Members"
        Private _SectionType As SectionTypes
        Private _ID As String
        Private _Style As SectionStyle
        Private _Left As Unit
        Private _Top As Unit
        Private _Visible As Boolean = True

#End Region

        Public ReadOnly Property SectionType() As SectionTypes
            Get
                Return _SectionType
            End Get
        End Property

        Public Property ID() As String
            Get
                If _ID Is Nothing Then
                    Return SectionType.ToString()
                Else
                    Return _ID
                End If
            End Get
            Set(ByVal Value As String)
                _ID = Value
            End Set
        End Property

        Public Overridable ReadOnly Property Style() As SectionStyle
            Get
                If Not StyleCreated Then
                    _Style = CreateStyle()
                End If
                Return _Style
            End Get
        End Property

        Public ReadOnly Property StyleCreated() As Boolean
            Get
                Return Not _Style Is Nothing
            End Get
        End Property

        Public Property BackColor() As System.Drawing.Color
            Get
                If StyleCreated Then
                    Return Style.BackColor
                Else
                    Return System.Drawing.Color.Empty
                End If
            End Get
            Set(ByVal Value As System.Drawing.Color)
                Style.BackColor = Value
            End Set
        End Property

        Public Property BorderColor() As System.Drawing.Color
            Get
                If StyleCreated Then
                    Return Style.Border.Color
                Else
                    Return System.Drawing.Color.Empty
                End If
            End Get
            Set(ByVal Value As System.Drawing.Color)
                Style.Border.Color = Value
            End Set
        End Property

        Public Property BorderStyle() As System.Web.UI.WebControls.BorderStyle
            Get
                If StyleCreated Then
                    Return Style.Border.Style
                Else
                    Return System.Web.UI.WebControls.BorderStyle.NotSet
                End If
            End Get
            Set(ByVal Value As System.Web.UI.WebControls.BorderStyle)
                Style.Border.Style = Value
            End Set
        End Property

        Public Property BorderWidth() As Unit
            Get
                If StyleCreated Then
                    Return Style.Border.Width
                Else
                    Return Unit.Empty
                End If
            End Get
            Set(ByVal Value As Unit)
                Style.Border.Width = Value
            End Set
        End Property

        Public Property Leading() As Unit
            Get
                If StyleCreated Then
                    Return Style.Leading
                Else
                    Return Unit.Empty
                End If
            End Get
            Set(ByVal Value As Unit)
                Style.Leading = Value
            End Set
        End Property

        Public ReadOnly Property Padding() As QuadUnit
            Get
                Return Style.Padding
            End Get
        End Property

        Public ReadOnly Property Font() As PDF.FontInfo
            Get
                Return Style.Font
            End Get
        End Property

        Public Property Height() As Unit
            Get
                If StyleCreated Then
                    Return Style.Height
                Else
                    Return Unit.Empty
                End If
            End Get
            Set(ByVal Value As Unit)
                Style.Height = Value
            End Set
        End Property

        Public Property HorizontalAlign() As System.Web.UI.WebControls.HorizontalAlign
            Get
                If StyleCreated Then
                    Return Style.HorizontalAlign
                Else
                    Return System.Web.UI.WebControls.HorizontalAlign.NotSet
                End If
            End Get
            Set(ByVal Value As System.Web.UI.WebControls.HorizontalAlign)
                Style.HorizontalAlign = Value
            End Set
        End Property

        Public Property Left() As Unit
            Get
                Return _Left
            End Get
            Set(ByVal Value As Unit)
                _Left = Value
            End Set
        End Property

        Public Property SpacingAfter() As Unit
            Get
                If StyleCreated Then
                    Return (Style.SpacingAfter)
                Else
                    Return Unit.Empty
                End If
            End Get
            Set(ByVal Value As Unit)
                Style.SpacingAfter = Value
            End Set
        End Property

        Public Property SpacingBefore() As Unit
            Get
                If StyleCreated Then
                    Return (Style.SpacingBefore)
                Else
                    Return Unit.Empty
                End If
            End Get
            Set(ByVal Value As Unit)
                Style.SpacingBefore = Value
            End Set
        End Property

        Public Property Top() As Unit
            Get
                Return _Top
            End Get
            Set(ByVal Value As Unit)
                _Top = Value
            End Set
        End Property

        Public Property Visible() As Boolean
            Get
                Return _Visible
            End Get
            Set(ByVal Value As Boolean)
                _Visible = Value
            End Set
        End Property

        Public Property Width() As Unit
            Get
                If StyleCreated Then
                    Return Style.Width
                Else
                    Return Unit.Empty
                End If
            End Get
            Set(ByVal Value As Unit)
                Style.Width = Value
            End Set
        End Property

#Region "Constructors"

        Public Sub New()
            'Default to undefined section type
        End Sub
        Public Sub New(ByVal SectionType As SectionTypes)
            _SectionType = SectionType
        End Sub

        Public Sub New(ByVal SectionType As SectionTypes, ByVal Node As XmlNode)
            Me.New(SectionType)
            ParseSectionAttributes(Node)
            ParseSectionStyle(Node)
        End Sub
#End Region

#Region "Private/Protected Methods"
        Protected Overridable Function CreateStyle() As SectionStyle
            Return New SectionStyle
        End Function

        Protected Sub ParseSectionAttributes(ByVal Node As XmlNode)
            Common.SetSubProperties(Me, "", Node)
        End Sub

        Protected Sub ParseSectionStyle(ByVal Node As XmlNode)
            Dim StyleNode As XmlNode = Node.Item("Style")
            If Not StyleNode Is Nothing Then
                Common.SetSubProperties(Me, "Style", StyleNode)
            End If
        End Sub
#End Region

#Region "Public Methods"
        Public Overridable Sub Render(ByVal writer As PDFWriter)
            If Visible Then
                Dim Paragraph As New iTextSharp.text.Paragraph
                With Paragraph
                    SpacingBefore.SetPoints(.SpacingBefore)
                    SpacingAfter.SetPoints(.SpacingBefore)
                    Common.SetHorizontalAlignment(.Alignment, Style.HorizontalAlign)
                    Padding.Left.SetPoints(.IndentationLeft)
                    Padding.Right.SetPoints(.IndentationRight)
                    If Not Leading.IsEmpty Then .Leading = Leading.ToPoints()
                End With
                Dim Content As iTextSharp.text.IElement = RenderedContent(writer)
                If Not Content Is Nothing Then
                    Paragraph.Add(Content)
                    If Not Style.Border.IsEmpty Then
                        Dim BorderChunk As New iTextSharp.text.Chunk("")
                        BorderChunk.SetGenericTag("PARAGRAPHBORDER " & Style.Border.ToString())
                        Paragraph.Add(BorderChunk)
                    End If
                    writer.AddToDocument(Paragraph)
                End If
            End If
        End Sub

        Public MustOverride Function RenderedContent(ByVal writer As PDFWriter) As iTextSharp.text.IElement

        Public Overridable Sub DataBind()
            'Derived classes will override this method when necessary to perform databinding
        End Sub

#End Region

    End Class
End Namespace
